class TabelineError(Exception):
    pass


class NonexistentColumnError(TabelineError):
    def __init__(self, column: str):
        super().__init__(column)
        self.column = column

    def __str__(self) -> str:
        return f"Cannot perform this operation using nonexistent column {self.column}"


class DuplicateColumnError(TabelineError):
    def __init__(self, column: str):
        super().__init__(column)
        self.column = column

    def __str__(self) -> str:
        return f"Column {self.column} is duplicated."


class RenameExistingError(TabelineError):
    def __init__(self, old_column: str, new_column: str):
        super().__init__(old_column, new_column)
        self.old_column = old_column
        self.new_column = new_column

    def __str__(self) -> str:
        return (
            f"Cannot rename {self.old_column} to {self.new_column} because"
            f" {self.new_column} already exists"
        )


class HasGroupsError(TabelineError):
    def __str__(self) -> str:
        return "Cannot perform this operation on a data frame with any group levels"


class NoGroupsError(TabelineError):
    def __str__(self) -> str:
        return "Cannot perform this operation on a data frame with no group levels"


class GroupColumnError(TabelineError):
    def __init__(self, column: str):
        super().__init__(column)
        self.column = column

    def __str__(self) -> str:
        return f"Cannot perform this operation on group column {self.column}"


class RowlessDataFrameError(TabelineError):
    def __str__(self) -> str:
        return "Cannot perform this operation on a data frame with no rows"


class IndexOutOfRangeError(TabelineError):
    def __init__(self, index: int, length: int, *, one_indexed: bool):
        self.index = index
        self.length = length
        self.one_indexed = one_indexed

    def __str__(self) -> str:
        base_string = "one" if self.one_indexed else "zero"
        return f"Cannot get index {self.index} ({base_string}-based) from data frame with {self.length} rows"


class UnmatchedColumnsError(TabelineError):
    def __init__(self, expected_columns: tuple[str, ...], actual_columns: tuple[str, ...]):
        super().__init__(expected_columns, actual_columns)
        self.expected_columns = expected_columns
        self.actual_columns = actual_columns

    def __str__(self) -> str:
        return (
            "Columns do not match\n"
            f"Expected columns: {list(self.expected_columns)}\n"
            f"Actual columns: {list(self.actual_columns)}"
        )


class UnmatchedGroupLevelsError(TabelineError):
    def __init__(
        self,
        expected_group_levels: tuple[tuple[str, ...], ...],
        actual_group_levels: tuple[tuple[str, ...], ...],
    ):
        super().__init__(expected_group_levels, actual_group_levels)
        self.expected_group_levels = expected_group_levels
        self.actual_group_levels = actual_group_levels

    def __str__(self) -> str:
        expected_str = ",".join(
            "[" + ",".join(level) + "]" for level in self.expected_group_levels
        )
        actual_str = ",".join("[" + ",".join(level) + "]" for level in self.actual_group_levels)
        return f"Group levels do not match; expected {expected_str} but got {actual_str}"


class UnmatchedHeightError(TabelineError):
    def __init__(self, expected_height: int, actual_height: int):
        super().__init__(expected_height, actual_height)
        self.expected_height = expected_height
        self.actual_height = actual_height

    def __str__(self) -> str:
        return (
            f"Heights do not match; expected {self.expected_height} but got {self.actual_height}"
        )


class NotSameError(TabelineError):
    def __init__(self, column: tuple[object, ...]):
        super().__init__(column)
        self.column = column

    def __str__(self) -> str:
        used = set()
        unique_elements = [x for x in self.column if x not in used and (used.add(x) or True)]
        return f"Assertion that all elements are the same failed, found {unique_elements}"



__all__ = ["concatenate_rows", "concatenate_columns"]

import polars as pl

from tabeline._data_frame import DataFrame
from tabeline.exceptions import (
    DuplicateColumnError,
    HasGroupsError,
    UnmatchedColumnsError,
    UnmatchedGroupLevelsError,
    UnmatchedHeightError,
)


def concatenate_rows(df: DataFrame, *dfs: DataFrame) -> DataFrame:
    for df_i in dfs:
        if df_i.column_names != df.column_names:
            raise UnmatchedColumnsError(df.column_names, df_i.column_names)

        if df.group_levels != df_i.group_levels:
            raise UnmatchedGroupLevelsError(df.group_levels, df_i.group_levels)

    return DataFrame(
        pl.concat([df._df, *(df_i._df for df_i in dfs)], how="vertical"), df.group_levels
    )


def concatenate_columns(df: DataFrame, *dfs: DataFrame) -> DataFrame:
    if df.group_levels != ():
        raise HasGroupsError()

    found_columns = set(df.column_names)
    for df_i in dfs:
        if df_i.height != df.height:
            raise UnmatchedHeightError(df.height, df_i.height)

        if df_i.group_levels != ():
            raise HasGroupsError()

        for column in df_i.column_names:
            if column in found_columns:
                raise DuplicateColumnError(column)
        found_columns.update(df_i.column_names)

    return DataFrame(
        pl.concat([df._df, *(df_i._df for df_i in dfs)], how="horizontal"), df.group_levels
    )



from __future__ import annotations

__all__ = ["DataFrame"]

from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Optional, Sequence, Union, overload

import polars as pl

from ._array import Array
from ._dummy import dummy_frame, dummy_name
from ._expression import substitute, to_polars
from ._expression.ast import Expression
from ._record import Record
from ._validation import assert_legal_columns, missing
from .exceptions import (
    GroupColumnError,
    HasGroupsError,
    IndexOutOfRangeError,
    NoGroupsError,
    NonexistentColumnError,
    RenameExistingError,
)

# Singleton indicating that the default value is to error
error = object()

if TYPE_CHECKING:
    import pandas as pd


class DataFrame:
    @overload
    def __init__(self, **columns: list[bool] | list[int] | list[float] | list[str]):
        pass

    @overload
    def __init__(
        self,
        polars_df: pl.DataFrame,
        group_levels: tuple[tuple[str, ...], ...] = (),
        height: Optional[int] = None,
    ):
        pass

    def __init__(
        self,
        polars_df: pl.DataFrame = missing,
        group_levels: tuple[tuple[str, ...], ...] = (),
        height: Optional[int] = None,
        /,
        **columns: list[bool] | list[int] | list[float] | list[str],
    ):
        self.group_levels: tuple[tuple[str, ...], ...]

        if polars_df is missing:
            self._df = pl.DataFrame(columns)
            self.group_levels = ()
            self.height = self._df.height  # With no columns, height is assumed to be 0
        elif len(columns) == 0:
            self._df = polars_df
            self.group_levels = group_levels
            if height is not None:
                self.height = height
            else:
                self.height = self._df.height  # With no height, height is taken from polars
        else:
            raise TypeError("DataFrame() takes 0 positional arguments but 1 was given")

    @staticmethod
    def columnless(*, height: int) -> DataFrame:
        return DataFrame(pl.DataFrame({}), (), height)

    @property
    def width(self) -> int:
        return self._df.width

    @property
    def shape(self) -> tuple[int, int]:
        return self.height, self.width

    @property
    def column_names(self) -> tuple[str, ...]:
        return tuple(self._df.columns)

    @property
    def group_names(self) -> tuple[str, ...]:
        return tuple(names for level in self.group_levels for names in level)

    @staticmethod
    def from_dict(columns: dict[str, Sequence[Any]], /) -> DataFrame:
        df = pl.from_dict(columns)
        return DataFrame(df)

    def to_dict(self) -> dict[str, Array]:
        if len(self.group_levels) != 0:
            raise HasGroupsError()

        return {name: Array(series) for name, series in self._df.to_dict().items()}

    @staticmethod
    def from_pandas(df: pd.DataFrame, /) -> DataFrame:
        if df.shape[1] == 0:
            # Polars does not understand columnless data frames
            return DataFrame.columnless(height=df.shape[0])

        return DataFrame(pl.from_pandas(df))

    def to_pandas(self) -> pd.DataFrame:
        import numpy as np
        import pandas as pd

        if len(self.group_levels) != 0:
            raise HasGroupsError()

        if self.width == 0:
            # Polars does not understand columnless data frames
            return pd.DataFrame(np.empty((self.height, 0)))

        return self._df.to_pandas()

    @staticmethod
    def from_polars(df: pl.DataFrame, /) -> DataFrame:
        return DataFrame(df)

    def to_polars(self) -> pl.DataFrame:
        if len(self.group_levels) != 0:
            raise HasGroupsError()

        return self._df

    @staticmethod
    def read_csv(path: Path, /) -> DataFrame:
        df = pl.read_csv(str(path))
        return DataFrame(df)

    def write_csv(self, path: Path, /) -> None:
        if len(self.group_levels) != 0:
            raise HasGroupsError()

        self._df.write_csv(str(path))

    def slice0(self, indexes: list[int], /) -> DataFrame:
        # Negative indexes don't make sense with on-based indexing, so they are
        # disallowed everywhere. This only catches non-positive indexes. Indexes
        # too large will be caught by Polars.
        for index in indexes:
            if index >= self.height or index < 0:
                raise IndexOutOfRangeError(index, self.height, one_indexed=False)

        if self.width == 0:
            for index in indexes:
                if index >= self.height:
                    raise IndexOutOfRangeError(index, self.height, one_indexed=False)
            return DataFrame(self._df, self.group_levels, len(indexes))
        else:
            group_names = self.group_names
            if len(group_names) == 0:
                # Polars chokes on empty groups
                return DataFrame(self._df.select(pl.all().gather(indexes)), self.group_levels)
            else:
                # There is no easy way to slice by groups in Polars. Using
                # `take` on a group causes the sliced columns to be a column of
                # lists that must be exploded at the end. The exploding results
                # in clustering, so getting back the original row order requires
                # tagging each row with an index and sorting the result by that.
                # https://stackoverflow.com/q/71373783/
                group_set = set(self.group_names)
                non_group_columns = [
                    column for column in self.column_names if column not in group_set
                ] + ["_index"]

                if len(indexes) == 0:
                    # Polars is not type stable when exploding no elements on a
                    # grouped data frame, so just drop all rows.
                    # https://github.com/pola-rs/polars/issues/6723
                    new_df = self._df.select(pl.all().gather(indexes))
                else:
                    new_df = (
                        self._df.lazy()
                        .with_columns(pl.int_range(0, pl.len()).alias("_index"))
                        .group_by(list(group_names), maintain_order=True)
                        .agg(pl.all().gather(indexes))
                        .explode(non_group_columns)
                        .sort("_index")
                        .drop("_index")
                        .collect()
                    )

                return DataFrame(new_df, self.group_levels)

    def slice1(self, indexes: list[int], /) -> DataFrame:
        # Negative indexes don't make sense with on-based indexing, so they are
        # disallowed everywhere. This only catches non-positive indexes. Indexes
        # too large will be caught by Polars.
        for index in indexes:
            if index <= 0:
                raise IndexOutOfRangeError(index, self.height, one_indexed=True)

        return self.slice0([index - 1 for index in indexes])

    def filter(self, predicate: str, /) -> DataFrame:
        expression = to_polars(Expression.parse(predicate).unwrap())

        if self.width == 0:
            # There is no way to evaluate an expression outside a data frame, so
            # make a dummy data frame with the right number of rows.
            # eager=True needed because lazy expressions are not allowed in
            # the constructor
            # https://github.com/pola-rs/polars/issues/2983
            height = dummy_frame(self.height).filter(expression).height
            return DataFrame(pl.DataFrame({}), self.group_levels, height)
        else:
            flat_groups = list(self.group_names)
            if len(flat_groups) > 0 and self.height > 0:
                windowed_expression = expression.over(flat_groups)
            else:
                # Polars chokes on an empty window
                # Polars chokes on window over rowless data frame
                windowed_expression = expression

            return DataFrame(self._df.filter(windowed_expression), self.group_levels)

    def distinct(self, *columns: str) -> DataFrame:
        assert_legal_columns(columns, self.column_names)

        # Polars chokes on duplicate columns
        group_set = set(self.group_names)
        unmentioned_columns = tuple(column for column in columns if column not in group_set)
        all_columns = list(self.group_names + unmentioned_columns)
        if len(all_columns) == 0:
            if self.height == 0:
                # Polars chokes on an empty distinct subset
                return self
            else:
                return DataFrame(self._df.head(1), self.group_levels, 1)
        else:
            return DataFrame(
                self._df.unique(subset=all_columns, maintain_order=True), self.group_levels
            )

    def unique(self) -> DataFrame:
        if self.width == 0:
            return DataFrame(pl.DataFrame({}), self.group_levels, min(1, self.height))

        return DataFrame(self._df.unique(maintain_order=True), self.group_levels)

    def cluster(self, *columns: str) -> DataFrame:
        assert_legal_columns(columns, self.column_names, self.group_names)

        flat_groups = self.group_names

        all_columns = list(self.group_names + columns)

        if len(all_columns) == 0:
            # Polars chokes on empty window columns
            return self

        if len(flat_groups) == 0:
            # Polars chokes on empty window columns
            return DataFrame(
                self._df.lazy()
                .with_columns(pl.int_range(0, pl.len()).alias("_index"))
                .with_columns(pl.min("_index").over(all_columns))
                .select(pl.all().sort_by(["_index"]))
                .drop("_index")
                .collect(),
                self.group_levels,
                self.height,
            )
        else:
            return DataFrame(
                self._df.lazy()
                .with_columns(pl.int_range(0, pl.len()).alias("_index"))
                .with_columns(pl.min("_index").over(all_columns))
                .select(pl.all().sort_by(["_index"]).over(list(self.group_names)))
                .drop("_index")
                .collect(),
                self.group_levels,
                self.height,
            )

    def sort(self, *columns: str) -> DataFrame:
        assert_legal_columns(columns, self.column_names, self.group_names)

        flat_groups = self.group_names
        if len(flat_groups) == 0 or self.height == 0:
            # Polars chokes on empty groups
            # Polars chokes on windowing over empty rows
            return DataFrame(self._df.sort(list(columns)), self.group_levels, self.height)
        else:
            return DataFrame(
                self._df.select(pl.all().sort_by(list(columns)).over(list(flat_groups))),
                self.group_levels,
                self.height,
            )

    def select(self, *columns: str) -> DataFrame:
        assert_legal_columns(columns, self.column_names)

        # Group columns cannot be deselected. Prepend any group columns that
        # were not explicitly selected so that group columns can be reordered
        # by select, but not lost. Unmentioned group columns are retained in
        # their original order.
        select_set = set(columns)
        group_set = set(self.group_names)
        unmentioned_groups = tuple(
            column
            for column in self.column_names
            if column in group_set and column not in select_set
        )

        return DataFrame(
            self._df.select(unmentioned_groups + columns), self.group_levels, self.height
        )

    def deselect(self, *columns: str) -> DataFrame:
        assert_legal_columns(columns, self.column_names, self.group_names)

        return DataFrame(self._df.drop(list(columns)), self.group_levels, self.height)

    def rename(self, **names: str) -> DataFrame:
        # Renames are processed simultaneously. Overwriting an existing column
        # is not permitted with this operation. Temporary column names may not
        # be used; columns to be swapped should simply be swapped. Renaming a
        # group column is legal.
        # Note that in dplyr, the order is new_name=old_name and in polars
        # old_name=new_name.
        # Sometime in the 0.13.x series, Polars switched from sequential to
        # parallel. dplyr has always been sequential.

        columns_set = set(self.column_names)

        for old_column in names.values():
            if old_column not in columns_set:
                raise NonexistentColumnError(old_column)
            else:
                columns_set.remove(old_column)

        for new_column in names.keys():
            if new_column in columns_set:
                raise RenameExistingError(names[new_column], new_column)
            else:
                columns_set.add(new_column)

        rename_map = {old: new for new, old in names.items()}

        groups = tuple(
            tuple(rename_map.get(column, column) for column in level)
            for level in self.group_levels
        )

        return DataFrame(self._df.rename(rename_map), groups, self.height)

    def mutate(self, **mutators: str) -> DataFrame:
        group_set = set(self.group_names)
        for column in mutators.keys():
            if column in group_set:
                raise GroupColumnError(column)

        if self.width == 0:
            df = dummy_frame(self.height)
        else:
            df = self._df

        # Both mutate and transmute must compute each column individually
        # because a column may refer to previous columns, which is not allowed
        # in a select context in polars:
        # https://stackoverflow.com/q/71105136/
        polars_operation = df.lazy()
        if len(group_set) == 0:
            for name, mutator in mutators.items():
                polars_operation = polars_operation.with_columns(
                    to_polars(Expression.parse(mutator).unwrap()).alias(name)
                )
        else:
            for name, mutator in mutators.items():
                polars_operation = polars_operation.with_columns(
                    to_polars(Expression.parse(mutator).unwrap())
                    .over(list(self.group_names))
                    .alias(name)
                )

        if self.width == 0:
            polars_operation = polars_operation.drop(dummy_name)

        return DataFrame(polars_operation.collect(), self.group_levels, self.height)

    def transmute(self, **mutators: str) -> DataFrame:
        mutated_df = self.mutate(**mutators)

        # Keep group columns and explicitly mentioned columns, all in the order
        # of group columns followed by explicit columns
        return mutated_df.select(*self.group_names, *mutators.keys())

    def group_by(
        self, *columns: str, order: Literal["original", "cluster", "sort"] = "original"
    ) -> DataFrame:
        assert_legal_columns(columns, self.column_names, self.group_names)

        if order == "original":
            ordered_df = self
        elif order == "cluster":
            ordered_df = self.cluster(*columns)
        elif order == "sort":
            ordered_df = self.sort(*columns)
        else:
            raise TypeError(
                f"For order, expected 'original', 'cluster', or 'sort', but got {order!r}"
            )

        return DataFrame(ordered_df._df, (*self.group_levels, columns), self.height)

    def group(
        self, *columns: str, order: Literal["original", "cluster", "sort"] = "original"
    ) -> DataFrame:
        return self.group_by(*columns, order=order)

    def ungroup(self) -> DataFrame:
        if len(self.group_levels) == 0:
            raise NoGroupsError()
        else:
            return DataFrame(self._df, self.group_levels[:-1], self.height)

    def summarize(self, **reducers: str) -> DataFrame:
        if len(self.group_levels) == 0:
            raise NoGroupsError()

        # There is no way to sequentially evaluate expressions in a group_by
        # context, so each reducer must be substituted into subsequent reducers:
        # https://stackoverflow.com/q/71120396/
        substituted_reducers = {}
        for name, reducer in reducers.items():
            substituted_reducers[name] = substitute(
                Expression.parse(reducer).unwrap(), substituted_reducers
            )

        polars_expressions = [
            to_polars(reducer).alias(name) for name, reducer in substituted_reducers.items()
        ]

        if len(polars_expressions) == 0:
            # Polars chokes on empty agg. This is equivalent to dropping
            # everything but the group columns and then running distinct.
            return self.select().distinct().ungroup()
        else:
            if self.width == 0:
                df = dummy_frame(self.height)
            else:
                df = self._df

            summarized = df.group_by(list(self.group_names), maintain_order=True).agg(
                polars_expressions
            )

            return DataFrame(summarized, self.group_levels[:-1])

    def spread(self, key: str, value: str, *, fill: Any = error) -> DataFrame:
        if len(self.group_levels) == 0:
            raise NoGroupsError()

        assert_legal_columns([key, value], self.column_names, self.group_names)

        df = self._df.pivot(
            index=list(self.group_names), on=key, values=value, aggregate_function=None
        )

        return DataFrame(df, self.group_levels[:-1])

    def gather(self, key: str, value: str, *columns: str) -> DataFrame:
        column_set = set(columns)
        all_columns = [column for column in self.column_names if column not in column_set]

        df = self._df.unpivot(
            index=all_columns, on=list(columns), variable_name=key, value_name=value
        )

        return DataFrame(df, self.group_levels).group_by(key)

    def _resolve_join_by(
        self,
        by: Optional[Sequence[Union[str, tuple[str, str]]]],
        right_column_names: Sequence[str],
    ) -> tuple[Optional[list[str]], Optional[list[str]]]:
        if by is not None:
            left_key_columns = []
            right_key_columns = []
            for by_i in by:
                if isinstance(by_i, str):
                    left_key_columns.append(by_i)
                    right_key_columns.append(by_i)
                else:
                    left_key_column, right_key_column = by_i
                    left_key_columns.append(left_key_column)
                    right_key_columns.append(right_key_column)

            assert_legal_columns(left_key_columns, self.column_names)
            assert_legal_columns(right_key_columns, right_column_names)
        else:
            right_columns_set = set(right_column_names)
            common_columns = [
                column for column in self.column_names if column in right_columns_set
            ]

            left_key_columns = common_columns
            right_key_columns = common_columns

        return left_key_columns, right_key_columns

    def inner_join(
        self,
        other: DataFrame,
        /,
        *,
        by: Optional[Sequence[Union[str, tuple[str, str]]]] = None,
    ) -> DataFrame:
        if len(self.group_levels) != 0 or len(other.group_levels) != 0:
            raise NotImplementedError(
                "Joins using grouped data frames is not currently implemented"
            )

        left_key_columns, right_key_columns = self._resolve_join_by(by, other.column_names)

        joined_df = (
            # Polars does not order the output, so sort by original orders
            self._df.lazy()
            .with_columns(pl.int_range(0, pl.len()).alias("_index1"))
            .join(
                other._df.lazy().with_columns(pl.int_range(0, pl.len()).alias("_index2")),
                left_on=left_key_columns,
                right_on=right_key_columns,
                how="inner",
            )
            .sort(["_index1", "_index2"])
            .drop(["_index1", "_index2"])
            .collect()
        )

        return DataFrame(joined_df)

    def left_join(
        self,
        other: DataFrame,
        /,
        *,
        by: Optional[Sequence[Union[str, tuple[str, str]]]] = None,
    ) -> DataFrame:
        if len(self.group_levels) != 0 or len(other.group_levels) != 0:
            raise NotImplementedError(
                "Joins using grouped data frames is not currently implemented"
            )

        left_key_columns, right_key_columns = self._resolve_join_by(by, other.column_names)

        joined_df = (
            # Polars does not order the output, so sort by original orders
            self._df.lazy()
            .with_columns(pl.int_range(0, pl.len()).alias("_index1"))
            .join(
                other._df.lazy().with_columns(pl.int_range(0, pl.len()).alias("_index2")),
                left_on=left_key_columns,
                right_on=right_key_columns,
                coalesce=True,
                how="left",
            )
            .sort(["_index1", "_index2"])
            .drop(["_index1", "_index2"])
            .collect()
        )
        return DataFrame(
            joined_df,
        )

    def outer_join(
        self,
        other: DataFrame,
        /,
        *,
        by: Optional[Sequence[Union[str, tuple[str, str]]]] = None,
    ) -> DataFrame:
        if len(self.group_levels) != 0 or len(other.group_levels) != 0:
            raise NotImplementedError(
                "Joins using grouped data frames is not currently implemented"
            )

        left_key_columns, right_key_columns = self._resolve_join_by(by, other.column_names)

        joined_df = (
            # Polars does not order the output, so sort by original orders
            self._df.lazy()
            .with_columns(pl.int_range(0, pl.len()).alias("_index1"))
            .join(
                other._df.lazy().with_columns(pl.int_range(0, pl.len()).alias("_index2")),
                left_on=left_key_columns,
                right_on=right_key_columns,
                coalesce=True,
                how="full",
            )
            .sort(["_index1", "_index2"])
            .drop(["_index1", "_index2"])
            .collect()
        )
        return DataFrame(
            joined_df,
        )

    @overload
    def __getitem__(self, key: tuple[int, str]) -> bool | int | float | str:
        pass

    @overload
    def __getitem__(self, key: tuple[int, Sequence[str]]) -> Record:
        pass

    @overload
    def __getitem__(self, key: tuple[Sequence[int] | Sequence[bool], str]) -> Array:
        pass

    @overload
    def __getitem__(self, key: tuple[Sequence[int] | Sequence[bool], Sequence[str]]) -> DataFrame:
        pass

    def __getitem__(
        self, key: tuple[int | Sequence[int] | Sequence[bool], str | Sequence[str]]
    ) -> DataFrame | Array | Record | bool | int | float | str:
        row_index, column_index = key

        if isinstance(column_index, str):
            if isinstance(row_index, int):
                return self._df.item(row_index, column_index)
            else:
                return Array(self._df[column_index][row_index])
        elif column_index == slice(None):
            if isinstance(row_index, int):
                return Record(self._df.row(row_index, named=True))
            else:
                return DataFrame(self._df[row_index, :], (), len(row_index))
        else:
            if isinstance(row_index, int):
                return Record(self._df.select(column_index).row(row_index, named=True))
            else:
                return DataFrame(self._df[row_index, column_index], (), len(row_index))

    def __eq__(self, other):
        if isinstance(other, DataFrame):
            if self.group_levels != other.group_levels:
                return False

            if self.width == 0:
                # Polars does not understand columnless data frames
                return self.height == other.height

            if self.height == 0:
                # Empty columns with different types are not equal in Polars,
                # but are in Tabeline.
                return self.column_names == other.column_names

            return self._df.equals(other._df)

        return NotImplemented

    def __repr__(self):
        column_strs = [f"{name} = {list(values)!r}" for name, values in self._df.to_dict().items()]
        group_strs = [f".group_by({', '.join(map(repr, level))})" for level in self.group_levels]
        return f"DataFrame({', '.join(column_strs)}){''.join(group_strs)}"

    def __str__(self):
        if len(self.group_levels) == 0:
            return str(self._df)
        else:
            groups_str = ",".join("[" + ",".join(level) + "]" for level in self.group_levels)
            return f"group levels: {groups_str}\n{self._df}"



from __future__ import annotations

__all__ = ["DataType"]

from enum import Enum, auto

import polars as pl


class DataType(Enum):
    Boolean = auto()
    Integer8 = auto()
    Integer16 = auto()
    Integer32 = auto()
    Integer64 = auto()
    Whole8 = auto()
    Whole16 = auto()
    Whole32 = auto()
    Whole64 = auto()
    Float32 = auto()
    Float64 = auto()
    String = auto()
    Nothing = auto()

    @staticmethod
    def from_polars(data_type: pl.DataType, /) -> DataType:
        if data_type == pl.Boolean:
            return DataType.Boolean
        elif data_type == pl.Int8:
            return DataType.Integer8
        elif data_type == pl.Int16:
            return DataType.Integer16
        elif data_type == pl.Int32:
            return DataType.Integer32
        elif data_type == pl.Int64:
            return DataType.Integer64
        elif data_type == pl.UInt8:
            return DataType.Whole8
        elif data_type == pl.UInt16:
            return DataType.Whole16
        elif data_type == pl.UInt32:
            return DataType.Whole32
        elif data_type == pl.UInt64:
            return DataType.Whole64
        elif data_type == pl.Float32:
            return DataType.Float32
        elif data_type == pl.Float64:
            return DataType.Float64
        elif data_type == pl.Utf8:
            return DataType.String
        elif data_type == pl.Null:
            return DataType.Nothing
        else:
            raise TypeError(f"Unsupported Polars data type: {data_type}")

    def to_polars(self) -> pl.DataType:
        if self == DataType.Boolean:
            return pl.Boolean()
        elif self == DataType.Integer8:
            return pl.Int8()
        elif self == DataType.Integer16:
            return pl.Int16()
        elif self == DataType.Integer32:
            return pl.Int32()
        elif self == DataType.Integer64:
            return pl.Int64()
        elif self == DataType.Whole8:
            return pl.UInt8()
        elif self == DataType.Whole16:
            return pl.UInt16()
        elif self == DataType.Whole32:
            return pl.UInt32()
        elif self == DataType.Whole64:
            return pl.UInt64()
        elif self == DataType.Float32:
            return pl.Float32()
        elif self == DataType.Float64:
            return pl.Float64()
        elif self == DataType.String:
            return pl.Utf8()
        elif self == DataType.Nothing:
            return pl.Null()



__all__ = ["parse_expression"]

from functools import reduce
from typing import Literal, Union

from parsita import ParserContext, Result, lit, opt, pred, reg, rep, rep1sep, repsep
from parsita.util import constant, splat

from . import ast
from .ast import Expression

reserved_names = {"None", "True", "False", "inf", "nan"}


def make_exponent(first: Expression, maybe_second: Union[tuple[()], tuple[Expression]]):
    if len(maybe_second) == 0:
        return first
    else:
        return ast.Power(first, maybe_second[0])


def make_term(first: Expression, rest: list[tuple[Literal["*", "/"], Expression]]) -> Expression:
    value = first
    for op, factor in rest:
        if op == "*":
            value = ast.Multiply(value, factor)
        elif op == "/":
            value = ast.Divide(value, factor)
        elif op == "%":
            value = ast.Mod(value, factor)
        else:
            raise NotImplementedError
    return value


def make_numeric_expression(
    first: Expression, rest: list[tuple[Literal["+", "-"], Expression]]
) -> Expression:
    value = first
    for op, term in rest:
        if op == "+":
            value = ast.Add(value, term)
        elif op == "-":
            value = ast.Subtract(value, term)
        else:
            raise NotImplementedError
    return value


def make_comparison_expression(
    first: Expression, maybe_second: Union[tuple[()], tuple[Expression]]
) -> Expression:
    if len(maybe_second) == 0:
        return first
    else:
        op, second = maybe_second[0]
        if op == "==":
            return ast.Equal(first, second)
        elif op == "!=":
            return ast.NotEqual(first, second)
        elif op == ">=":
            return ast.GreaterThanOrEqual(first, second)
        elif op == "<=":
            return ast.LessThanOrEqual(first, second)
        elif op == ">":
            return ast.GreaterThan(first, second)
        elif op == "<":
            return ast.LessThan(first, second)
        else:
            raise NotImplementedError


class ExpressionParser(ParserContext, whitespace=r"[ ]*"):
    name = pred(reg(r"[A-Za-z_][A-Za-z_0-9]*"), lambda x: x not in reserved_names, "non-keyword")

    # Atoms
    null = lit("None") > constant(ast.NullLiteral())

    true = lit("True") > constant(ast.BooleanLiteral(True))
    false = lit("False") > constant(ast.BooleanLiteral(False))
    boolean_literal = true | false

    float_literal = reg(r"\d+((\.\d+([Ee][+-]?\d+)?)|((\.\d+)?[Ee][+-]?\d+))") | reg(
        r"\binf\b"
    ) | reg(r"\bnan\b") > (lambda x: ast.FloatLiteral(float(x)))

    integer_literal = reg(r"\d+") > (lambda x: ast.IntegerLiteral(int(x)))

    string_literal = reg(r"'[^']*'") > (lambda x: ast.StringLiteral(x[1:-1]))

    variable = name > ast.Variable

    function = name & "(" >> repsep(expression, ",") << ")" > splat(ast.Call)

    parentheses = "(" >> expression << ")"

    atom = (
        null
        | boolean_literal
        | float_literal
        | integer_literal
        | string_literal
        | function
        | variable
        | parentheses
    )

    # Numeric operators
    positive = "+" >> factor > ast.Positive
    negative = "-" >> factor > ast.Negative
    exponent = atom & opt("**" >> factor) > splat(make_exponent)
    factor = positive | negative | exponent

    term = factor & rep(lit("*", "/", "%") & factor) > splat(make_term)
    numeric_expression = term & rep(lit("+", "-") & term) > splat(make_numeric_expression)

    # Comparison operators
    comparison_expression = numeric_expression & opt(
        lit("==", "!=", ">=", "<=", ">", "<") & numeric_expression
    ) > splat(make_comparison_expression)

    # Boolean operators
    not_expression = "~" >> unary_boolean > ast.Not

    unary_boolean = not_expression | comparison_expression

    and_expression = rep1sep(unary_boolean, "&") > (lambda x: reduce(ast.And, x))
    expression = rep1sep(and_expression, "|") > (lambda x: reduce(ast.Or, x))


def parse_expression(text: str) -> Result:
    return ExpressionParser.expression.parse(text)



__all__ = ["to_polars"]

from functools import singledispatch

import polars as pl

from . import ast
from ._functions import function_by_name


@singledispatch
def to_polars(self: ast.Expression) -> pl.Expr:
    raise NotImplementedError()


@to_polars.register(ast.NullLiteral)
def to_polars_null_literal(self: ast.NullLiteral) -> pl.Expr:
    return pl.lit(None)


@to_polars.register(ast.BooleanLiteral)
def to_polars_boolean_literal(self: ast.BooleanLiteral) -> pl.Expr:
    return pl.lit(self.value, dtype=pl.datatypes.Boolean)


@to_polars.register(ast.IntegerLiteral)
def to_polars_integer_literal(self: ast.IntegerLiteral) -> pl.Expr:
    return pl.lit(self.value, dtype=pl.datatypes.Int64)


@to_polars.register(ast.FloatLiteral)
def to_polars_float_literal(self: ast.FloatLiteral) -> pl.Expr:
    return pl.lit(self.value, dtype=pl.datatypes.Float64)


@to_polars.register(ast.StringLiteral)
def to_polars_string_literal(self: ast.StringLiteral) -> pl.Expr:
    return pl.lit(self.value, dtype=pl.datatypes.Utf8)


@to_polars.register(ast.Variable)
def to_polars_variable(self: ast.Variable) -> pl.Expr:
    return pl.col(self.name)


@to_polars.register(ast.Positive)
def to_polars_positive(self: ast.Positive) -> pl.Expr:
    return to_polars(self.content)


@to_polars.register(ast.Negative)
def to_polars_negative(self: ast.Negative) -> pl.Expr:
    return -to_polars(self.content)


@to_polars.register(ast.Add)
def to_polars_add(self: ast.Add) -> pl.Expr:
    return to_polars(self.left) + to_polars(self.right)


@to_polars.register(ast.Subtract)
def to_polars_subtract(self: ast.Subtract) -> pl.Expr:
    return to_polars(self.left) - to_polars(self.right)


@to_polars.register(ast.Multiply)
def to_polars_multiply(self: ast.Multiply) -> pl.Expr:
    return to_polars(self.left) * to_polars(self.right)


@to_polars.register(ast.Divide)
def to_polars_divide(self: ast.Divide) -> pl.Expr:
    return to_polars(self.left) / to_polars(self.right)


@to_polars.register(ast.Mod)
def to_polars_mod(self: ast.Mod) -> pl.Expr:
    return to_polars(self.left) % to_polars(self.right)


@to_polars.register(ast.Power)
def to_polars_power(self: ast.Power) -> pl.Expr:
    return to_polars(self.left) ** to_polars(self.right)


@to_polars.register(ast.Call)
def to_polars_call(self: ast.Call) -> pl.Expr:
    function = function_by_name[self.name]
    return function.implementation(*map(to_polars, self.arguments))


@to_polars.register(ast.Equal)
def to_polars_equal(self: ast.Equal) -> pl.Expr:
    return to_polars(self.left) == to_polars(self.right)


@to_polars.register(ast.NotEqual)
def to_polars_not_equal(self: ast.NotEqual) -> pl.Expr:
    return to_polars(self.left) != to_polars(self.right)


@to_polars.register(ast.GreaterThanOrEqual)
def to_polars_greater_than_or_equal(self: ast.GreaterThanOrEqual) -> pl.Expr:
    return to_polars(self.left) >= to_polars(self.right)


@to_polars.register(ast.LessThanOrEqual)
def to_polars_less_than_or_equal(self: ast.LessThanOrEqual) -> pl.Expr:
    return to_polars(self.left) <= to_polars(self.right)


@to_polars.register(ast.GreaterThan)
def to_polars_greater_than(self: ast.GreaterThan) -> pl.Expr:
    return to_polars(self.left) > to_polars(self.right)


@to_polars.register(ast.LessThan)
def to_polars_less_than(self: ast.LessThan) -> pl.Expr:
    return to_polars(self.left) < to_polars(self.right)


@to_polars.register(ast.Not)
def to_polars_not(self: ast.Not) -> pl.Expr:
    return ~to_polars(self.content)


@to_polars.register(ast.And)
def to_polars_and(self: ast.And) -> pl.Expr:
    return to_polars(self.left) & to_polars(self.right)


@to_polars.register(ast.Or)
def to_polars_or(self: ast.Or) -> pl.Expr:
    return to_polars(self.left) | to_polars(self.right)



__all__ = ["missing", "assert_legal_columns"]

from typing import Sequence

from tabeline.exceptions import DuplicateColumnError, GroupColumnError, NonexistentColumnError

missing = object()


def assert_legal_columns(
    columns: Sequence[str],
    existing_columns: Sequence[str],
    illegal_group_columns: Sequence[str] = (),
) -> None:
    columns_set: set[str] = set()
    existing_columns_set = set(existing_columns)
    group_column_set = set(illegal_group_columns)
    for column in columns:
        if column in columns_set:
            raise DuplicateColumnError(column)

        if column not in existing_columns_set:
            raise NonexistentColumnError(column)

        if column in group_column_set:
            raise GroupColumnError(column)

        columns_set.add(column)



__all__ = ["substitute"]

from functools import singledispatch

from . import ast


@singledispatch
def substitute(self: ast.Expression, variables: dict[str, ast.Expression]) -> ast.Expression:
    raise NotImplementedError()


@substitute.register(ast.NullLiteral)
def substitute_null_literal(
    self: ast.Expression, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return self


@substitute.register(ast.BooleanLiteral)
def substitute_boolean_literal(
    self: ast.BooleanLiteral, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return self


@substitute.register(ast.IntegerLiteral)
def substitute_integer_literal(
    self: ast.IntegerLiteral, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return self


@substitute.register(ast.FloatLiteral)
def substitute_float_literal(
    self: ast.FloatLiteral, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return self


@substitute.register(ast.StringLiteral)
def substitute_string_literal(
    self: ast.StringLiteral, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return self


@substitute.register(ast.Variable)
def substitute_variable(
    self: ast.Variable, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return variables.get(self.name, self)


@substitute.register(ast.Positive)
def substitute_positive(
    self: ast.Positive, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return ast.Positive(substitute(self.content, variables))


@substitute.register(ast.Negative)
def substitute_negative(
    self: ast.Negative, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return ast.Negative(substitute(self.content, variables))


@substitute.register(ast.Add)
def substitute_add(self: ast.Add, variables: dict[str, ast.Expression]) -> ast.Expression:
    return ast.Add(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.Subtract)
def substitute_subtract(
    self: ast.Subtract, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return ast.Subtract(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.Multiply)
def substitute_multiply(
    self: ast.Multiply, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return ast.Multiply(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.Divide)
def substitute_divide(self: ast.Divide, variables: dict[str, ast.Expression]) -> ast.Expression:
    return ast.Divide(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.Mod)
def substitute_mod(self: ast.Mod, variables: dict[str, ast.Expression]) -> ast.Expression:
    return ast.Mod(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.Power)
def substitute_power(self: ast.Power, variables: dict[str, ast.Expression]) -> ast.Expression:
    return ast.Power(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.Call)
def substitute_call(self: ast.Call, variables: dict[str, ast.Expression]) -> ast.Expression:
    return ast.Call(self.name, [substitute(argument, variables) for argument in self.arguments])


@substitute.register(ast.Equal)
def substitute_equal(self: ast.Equal, variables: dict[str, ast.Expression]) -> ast.Expression:
    return ast.Equal(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.NotEqual)
def substitute_not_equal(
    self: ast.NotEqual, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return ast.NotEqual(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.GreaterThanOrEqual)
def substitute_greater_than_or_equal(
    self: ast.GreaterThanOrEqual, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return ast.GreaterThanOrEqual(
        substitute(self.left, variables), substitute(self.right, variables)
    )


@substitute.register(ast.LessThanOrEqual)
def substitute_less_than_or_equal(
    self: ast.LessThanOrEqual, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return ast.LessThanOrEqual(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.GreaterThan)
def substitute_greater_than(
    self: ast.GreaterThan, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return ast.GreaterThan(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.LessThan)
def substitute_less_than(
    self: ast.LessThan, variables: dict[str, ast.Expression]
) -> ast.Expression:
    return ast.LessThan(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.Not)
def substitute_not(self: ast.Not, variables: dict[str, ast.Expression]) -> ast.Expression:
    return ast.Not(substitute(self.content, variables))


@substitute.register(ast.And)
def substitute_and(self: ast.And, variables: dict[str, ast.Expression]) -> ast.Expression:
    return ast.And(substitute(self.left, variables), substitute(self.right, variables))


@substitute.register(ast.Or)
def substitute_or(self: ast.Or, variables: dict[str, ast.Expression]) -> ast.Expression:
    return ast.Or(substitute(self.left, variables), substitute(self.right, variables))



from polars.testing import assert_frame_equal

from ._data_frame import DataFrame


def assert_data_frame_equal(
    left: DataFrame, right: DataFrame, reltol: float = 0.0, abstol: float = 0.0
) -> None:
    assert left.shape == right.shape
    assert left.group_levels == right.group_levels
    assert_frame_equal(left._df, right._df, check_dtypes=False, rtol=reltol, atol=abstol)



from ._array import Array
from ._concatenate import concatenate_columns, concatenate_rows
from ._data_frame import DataFrame
from ._data_type import DataType
from ._record import Record

DataTable = DataFrame



from __future__ import annotations

__all__ = ["Record"]

from typing import Iterator, overload

from ._validation import missing


class Record:
    @overload
    def __init__(self, **items: bool | int | float | str | None):
        pass

    @overload
    def __init__(self, data: dict[str, bool | int | float | str | None], /):
        pass

    def __init__(self, data=missing, /, **items):
        if data is missing:
            self._data = items
        else:
            self._data = data

    @staticmethod
    def from_dict(data: dict[str, bool | int | float | str | None], /) -> Record:
        return Record(data)

    def to_dict(self) -> dict[str, bool | int | float | str | None]:
        return self._data

    def __eq__(self, other: object) -> bool:
        # Cannot add `strict=True` to `zip` and still run on Python 3.9
        if isinstance(other, Record):
            return len(self) == len(other) and all(
                item == other_item
                for item, other_item in zip(self._data.items(), other._data.items())
            )
        elif isinstance(other, dict):
            return len(self) == len(other) and all(
                item == other_item for item, other_item in zip(self._data.items(), other.items())
            )
        return NotImplemented

    def __len__(self) -> int:
        return len(self._data)

    def __getitem__(self, key: str) -> bool | int | float | str:
        return self._data[key]

    def items(self) -> Iterator[str, bool | int | float | str]:
        yield from self._data.items()

    def __str__(self) -> str:
        return str(self._data)

    def __repr__(self) -> str:
        return f"Record({', '.join(f'{key}={value!r}' for key, value in self._data.items())})"



from __future__ import annotations

__all__ = ["Array"]

from collections.abc import Sequence
from typing import Generic, Iterator, TypeVar, overload

import numpy as np
import polars as pl
from polars.exceptions import InvalidOperationError

from ._data_type import DataType

Element = TypeVar("Element", bound=DataType)


class Array(Sequence, Generic[Element]):
    @overload
    def __init__(self, *elements: bool | None, data_type: DataType | None = None) -> None:
        pass

    @overload
    def __init__(self, *elements: int | None, data_type: DataType | None = None) -> None:
        pass

    @overload
    def __init__(self, *elements: float | int | None, data_type: DataType | None = None) -> None:
        pass

    @overload
    def __init__(self, *elements: str | None, data_type: DataType | None = None) -> None:
        pass

    @overload
    def __init__(self, elements: pl.Array, /, *, data_type: DataType | None = None) -> None:
        pass

    def __init__(self, *elements, data_type: DataType | None = None) -> None:
        if len(elements) > 0 and isinstance(elements[0], pl.Series):
            if len(elements) != 1:
                raise TypeError(
                    "Passing a Series is the private constructor, "
                    "which requires exactly one argument"
                )
            if data_type is not None:
                raise TypeError(
                    "Passing a Series is the private constructor, "
                    "which does not allow setting the data type"
                )

            series = elements[0]

            # Extract data type from Polars dtype
            data_type = DataType.from_polars(series.dtype)
        else:
            data_type = coerce_default_element_type(default_element_type(elements), data_type)
            series = pl.Series(values=elements, dtype=data_type.to_polars())

        self.data_type = data_type
        self._series = series

    @staticmethod
    def from_sequence(
        sequence: Sequence[None | bool | int | float | str],
        /,
        *,
        data_type: DataType | None = None,
    ) -> Array:
        return Array(*sequence, data_type=data_type)

    @staticmethod
    def from_polars(array: pl.Array, /) -> Array:
        return Array(array)

    def to_polars(self) -> pl.Array:
        return self._series

    @staticmethod
    def from_numpy(array: np.ndarray, /) -> Array:
        return Array(pl.Series(values=array))

    def to_numpy(self) -> np.ndarray:
        return self._series.to_numpy()

    def __len__(self) -> int:
        return len(self._series)

    def __getitem__(self, index: int) -> bool | int | float | str:
        return self._series[index]

    def __iter__(self) -> Iterator[bool | int | float | str]:
        return iter(self._series)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, (Array, pl.Series)):
            # Series.series_equal does not consider NaNs equal, so we have to
            # try to find the NaNs and make those elements equal. But not all
            # data types support NaNs, so bail when is_nan fails.

            self_series = self.to_polars()

            if isinstance(other, Array):
                other_series = other.to_polars()
            else:
                other_series = other

            equal_elements = self_series.eq_missing(other_series)

            try:
                self_nan = self_series.is_nan().fill_null(False)
            except InvalidOperationError:
                return equal_elements.all()

            try:
                other_nan = other_series.is_nan().fill_null(False)
            except InvalidOperationError:
                return equal_elements.all()

            return (equal_elements | (self_nan & other_nan)).all()
        elif isinstance(other, np.ndarray):
            # Exit early on empty arrays to avoid NumPy/ctypes warning in Python 3.9
            if len(self) == 0 or len(other) == 0:
                return len(self) == len(other)

            return array_equal(self.to_numpy(), other)
        return NotImplemented

    def __str__(self) -> str:
        return "[\n" + "".join(f"    {x}" for x in self._series) + "\n]"

    def __repr__(self) -> str:
        return f"Array({', '.join(repr(x) for x in self._series)})"

    def __array__(self) -> np.ndarray:
        return self.to_numpy()

    def __class_getitem__(cls, data_type: DataType) -> SpecializedArray:
        return SpecializedArray(data_type)


class SpecializedArray:
    def __init__(self, data_type: DataType, /) -> None:
        self.data_type = data_type

    def __call__(self, *elements: object) -> Array:
        return Array(*elements, data_type=self.data_type)


def array_equal(left: np.ndarray, right: np.ndarray) -> bool:
    import numpy as np

    # equal_nan crashes on non-numeric arrays, catch that error and try again
    try:
        return np.array_equal(left, right, equal_nan=True)
    except TypeError:
        return np.array_equal(left, right)


def default_element_type(
    sequence: Sequence[bool | None]
    | Sequence[int | None]
    | Sequence[float | int | None]
    | Sequence[str | None],
) -> DataType.Boolean | DataType.Integer64 | DataType.Float64 | DataType.String | DataType.Nothing:
    # Only a subset of data types are possible when converting Python types
    best_type: (
        DataType.Boolean
        | DataType.Integer64
        | DataType.Float64
        | DataType.String
        | DataType.Nothing
    ) = DataType.Nothing

    for element in sequence:
        if best_type == DataType.Nothing:
            if element is None:
                pass
            elif isinstance(element, bool):
                best_type = DataType.Boolean
            elif isinstance(element, int):
                best_type = DataType.Integer64
            elif isinstance(element, float):
                best_type = DataType.Float64
            elif isinstance(element, str):
                best_type = DataType.String
            else:
                raise TypeError(
                    f"Expected bool, int, float, str, or None, but got {type(element)}: {element}"
                )
        elif best_type == DataType.Boolean:
            if element is None or isinstance(element, bool):
                pass
            else:
                raise TypeError(f"Expected bool, but got {type(element)}: {element}")
        elif best_type == DataType.Integer64:
            if element is None or isinstance(element, int):
                pass
            elif isinstance(element, float):
                best_type = DataType.Float64
            else:
                raise TypeError(f"Expected int, but got {type(element)}: {element}")
        elif best_type == DataType.Float64:
            if element is None or isinstance(element, int) or isinstance(element, float):
                pass
            else:
                raise TypeError(f"Expected int or float, but got {type(element)}: {element}")
        elif best_type == DataType.String:
            if element is None or isinstance(element, str):
                pass
            else:
                raise TypeError(f"Expected str, but got {type(element)}: {element}")

    return best_type


def coerce_default_element_type(
    default: DataType.Boolean
    | DataType.Integer64
    | DataType.Float64
    | DataType.String
    | DataType.Nothing,
    given: DataType | None,
) -> DataType:
    if given is None:
        return default
    elif default == DataType.Boolean and given == DataType.Boolean:
        return given
    elif default == DataType.Integer64 and given in [
        DataType.Integer8,
        DataType.Integer16,
        DataType.Integer32,
        DataType.Integer64,
        DataType.Whole8,
        DataType.Whole16,
        DataType.Whole32,
        DataType.Whole64,
        DataType.Float32,
        DataType.Float64,
    ]:
        return given
    elif default == DataType.Float64 and given in [DataType.Float32, DataType.Float64]:
        return given
    elif default == DataType.String and given == DataType.String:
        return given
    elif default == DataType.Nothing and given == DataType.Nothing:
        return given
    else:
        raise TypeError(f"Expected data_type to be compatible with {default}, but got {given}")



__all__ = ["dummy_name", "dummy_frame"]

import polars as pl

dummy_name = "_dummy"


def dummy_frame(height: int) -> pl.DataFrame:
    # Lots of things in Polars don't work if the DataFrame has no columns
    # This constructs an effectively columnless data frame with the correct height
    return pl.DataFrame({dummy_name: pl.repeat(False, height, eager=True)})



__all__ = ["Function", "function_by_name"]

from dataclasses import dataclass
from typing import Any, Callable, Generic, TypeVar, Union
from typing_extensions import ParamSpec  # Not present in Python 3.9

import numpy as np
import polars as pl

from ..exceptions import NotSameError

P = ParamSpec("P")
R = TypeVar("R")


@dataclass(frozen=True)
class Function(Generic[P, R]):
    name: str
    implementation: Callable[P, R]


def polars_same(args):
    x = args[0]
    if x.n_unique() == 0:
        return pl.Series([], dtype=x.dtype)
    elif x.n_unique() == 1:
        return pl.Series([x[0]], dtype=x.dtype)
    else:
        raise NotSameError(x.unique().to_list())


def has_null(x: Union[pl.Expr, float]):
    if isinstance(x, float):
        return False
    elif isinstance(x, pl.Series):
        return x.is_null().any()


def polars_interp(args):
    x, xp, fp = args

    if has_null(x) or has_null(xp) or has_null(fp):
        return pl.Series(values=[None], dtype=pl.Float64)

    if len(x) == 1:
        # Polars does not support scalars
        # Assume all length-1 vectors are scalars
        x = x[0]

    return pl.Series(values=[np.interp(x, xp, fp)], dtype=pl.Float64)


built_in_functions: list[Function[Any, Any]] = [
    # Constant
    Function("n", lambda: pl.len()),  # https://stackoverflow.com/a/71644903/1485877
    Function("row_index0", lambda: pl.int_range(0, pl.len())),
    Function("row_index1", lambda: pl.int_range(1, pl.len() + 1)),
    # Numeric -> numeric elementwise
    Function("abs", lambda x: x.abs()),
    Function("sqrt", lambda x: x.sqrt()),
    Function("log", lambda x: x.log()),
    Function("log2", lambda x: x.log(2)),
    Function("log10", lambda x: x.log10()),
    Function("exp", lambda x: x.exp()),
    Function("pow", lambda x, y: x**y),
    Function("sin", lambda x: x.sin()),
    Function("cos", lambda x: x.cos()),
    Function("tan", lambda x: x.tan()),
    Function("arcsin", lambda x: x.arcsin()),
    Function("arccos", lambda x: x.arccos()),
    Function("arctan", lambda x: x.arctan()),
    Function("floor", lambda x: x.floor()),
    Function("ceil", lambda x: x.ceil()),
    # Numeric -> boolean elementwise
    Function("is_null", lambda x: x.is_null()),
    Function("is_nan", lambda x: x.is_nan()),
    Function("is_finite", lambda x: x.is_finite()),
    # Casting elementwise
    Function("to_boolean", lambda x: x.cast(pl.Boolean)),
    Function("to_integer", lambda x: x.cast(pl.Int64)),
    Function("to_float", lambda x: x.cast(pl.Float64)),
    Function("to_string", lambda x: x.cast(pl.Utf8)),
    # Other elementwise
    Function(
        "if_else",
        lambda condition, true, false=None: pl.when(condition).then(true).otherwise(false),
    ),
    # Numeric -> numeric reduction
    Function("std", lambda x: pl.when(x.is_null().any()).then(None).otherwise(x.std())),
    Function("var", lambda x: pl.when(x.is_null().any()).then(None).otherwise(x.var())),
    Function("max", lambda x: pl.when(x.is_null().any()).then(None).otherwise(x.max())),
    Function("min", lambda x: pl.when(x.is_null().any()).then(None).otherwise(x.min())),
    Function("sum", lambda x: pl.when(x.is_null().any()).then(None).otherwise(x.sum())),
    Function("mean", lambda x: pl.when(x.is_null().any()).then(None).otherwise(x.mean())),
    Function("median", lambda x: pl.when(x.is_null().any()).then(None).otherwise(x.median())),
    Function(
        "quantile",
        lambda x, quantile: pl.when(x.is_null().any())
        .then(None)
        .otherwise(x.quantile(pl.select(quantile).item(), interpolation="linear")),
    ),  # https://stackoverflow.com/a/71721580/
    Function(
        "trapz",
        lambda x, y: pl.when(x.is_null().any().or_(y.is_null().any()))
        .then(None)
        .otherwise(0.5 * ((x - x.shift()) * (y + y.shift())).sum()),
    ),  # https://github.com/pola-rs/polars/issues/3043
    Function(
        "interp",
        lambda x, xp, fp: pl.map_groups(
            exprs=[x, xp, fp],
            function=polars_interp,
        ),
    ),  # https://stackoverflow.com/a/69585269/
    # Boolean -> boolean reduction
    Function("any", lambda x: x.any(ignore_nulls=False)),
    Function("all", lambda x: x.all(ignore_nulls=False)),
    # Any -> any reduction
    # Function("mode", lambda x: x.mode()),  # Not type stable  # noqa: ERA001
    Function("first", lambda x: x.first()),
    Function("last", lambda x: x.last()),
    Function("same", lambda x: pl.map_groups(exprs=[x], function=polars_same)),
]

function_by_name: dict[str, Function[Any, Any]] = {x.name: x for x in built_in_functions}



from __future__ import annotations

__all__ = [
    "Expression",
    "NullLiteral",
    "BooleanLiteral",
    "IntegerLiteral",
    "FloatLiteral",
    "StringLiteral",
    "Variable",
    "Positive",
    "Negative",
    "Add",
    "Subtract",
    "Multiply",
    "Divide",
    "Mod",
    "Power",
    "Call",
    "Equal",
    "NotEqual",
    "GreaterThan",
    "LessThan",
    "GreaterThanOrEqual",
    "LessThanOrEqual",
    "Not",
    "And",
    "Or",
]

from dataclasses import dataclass

from parsita import Result


class Expression:
    @staticmethod
    def parse(text: str) -> Result:
        from ._parser import parse_expression

        return parse_expression(text)


@dataclass(frozen=True)
class NullLiteral(Expression):
    pass


@dataclass(frozen=True)
class BooleanLiteral(Expression):
    value: bool


@dataclass(frozen=True)
class IntegerLiteral(Expression):
    value: int


@dataclass(frozen=True)
class FloatLiteral(Expression):
    value: float


@dataclass(frozen=True)
class StringLiteral(Expression):
    value: str


@dataclass(frozen=True)
class Variable(Expression):
    name: str


@dataclass(frozen=True)
class Positive(Expression):
    content: Expression


@dataclass(frozen=True)
class Negative(Expression):
    content: Expression


@dataclass(frozen=True)
class Add(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class Subtract(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class Multiply(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class Divide(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class Mod(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class Power(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class Call(Expression):
    name: str
    arguments: list[Expression]


@dataclass(frozen=True)
class Equal(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class NotEqual(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class GreaterThanOrEqual(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class LessThanOrEqual(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class GreaterThan(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class LessThan(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class Not(Expression):
    content: Expression


@dataclass(frozen=True)
class And(Expression):
    left: Expression
    right: Expression


@dataclass(frozen=True)
class Or(Expression):
    left: Expression
    right: Expression
